#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-11-10
# @Filename: tools.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import contextlib
import inspect
import json
import logging
import os
import pathlib
import subprocess
import time
from contextlib import suppress
from functools import partial, wraps
from logging import FileHandler, getLogger

from typing import TYPE_CHECKING, Any, Callable, Coroutine, Sequence

import httpx
from jinja2 import Environment, FileSystemLoader
from rich.console import Console
from rich.progress import BarColumn, MofNCompleteColumn, Progress, TaskID, TextColumn

from lvmopstools.clu import CluClient
from lvmopstools.retrier import Retrier
from sdsstools.logger import CustomJsonFormatter
from sdsstools.utils import run_in_executor

from lvmcryo.config import ParameterOrigin, get_internal_config


if TYPE_CHECKING:
    from datetime import datetime

    from sdsstools.logger import SDSSLogger

    from lvmcryo.config import Config
    from lvmcryo.handlers.ln2 import LN2Handler


def is_container():
    """Returns `True` if the code is running inside a container."""

    is_container = os.getenv("IS_CONTAINER", None)
    if not is_container or is_container in ["", "0"]:
        return False

    return True


class TimerProgressBar:
    """A progress bar with a timer."""

    def __init__(self, console: Console | None = None):
        self.progress = Progress(
            TextColumn("[yellow]({task.fields[label]})"),
            TextColumn("[progress.description]{task.description}"),
            BarColumn(bar_width=None),
            MofNCompleteColumn(),
            TextColumn("s"),
            refresh_per_second=1,
            expand=True,
            transient=False,
            auto_refresh=False,
            console=console,  # Need to use same console as logger.
        )
        self.console = self.progress.console

        self._tasks: dict[TaskID, asyncio.Task] = {}

    async def add_timer(
        self,
        max_time: float,
        label: str = "",
        initial_description: str = "Fill in progress ...",
        complete_description: str = "Fill complete",
    ):
        """Starts the timer."""

        task_id = self.progress.add_task(
            f"[blue] {initial_description} ",
            total=int(max_time),
            label=label,
        )

        self.progress.start()

        start_time = time.time()

        async def update_timer():
            elapsed = 0
            while True:
                if elapsed > max_time:
                    break

                self.progress.update(task_id, advance=1, visible=True)
                asyncio.create_task(run_in_executor(self.progress.refresh))

                await asyncio.sleep(1)
                elapsed += 1

        done_timer_partial = partial(
            self._done_timer,
            start_time,
            max_time,
            complete_description,
            task_id,
        )

        _task = asyncio.create_task(update_timer())
        _task.add_done_callback(done_timer_partial)

        self._tasks[task_id] = _task

        return task_id

    def _done_timer(
        self,
        start_time: float,
        max_time: float,
        complete_description: str,
        task_id: TaskID,
        *_,
    ):
        completed = int(time.time() - start_time)
        if completed > max_time:
            completed = int(max_time)

        self.progress.update(
            task_id,
            description=f"[green] {complete_description} ",
        )
        self.progress.refresh()

    async def stop_timer(self, task_id: TaskID, clear: bool = False):
        """Cancels a timer."""

        if task_id not in self._tasks:
            raise ValueError("Task ID not found.")

        _task = self._tasks[task_id]
        await cancel_task(_task)

        del self._tasks[task_id]

        if clear:
            self.progress.update(task_id, visible=False)

        await asyncio.sleep(1)  # Extra time for renders.

    def close(self):
        """Closes the progress bar."""

        self.progress.console.clear_live()
        self.progress.stop()


async def cancel_task(task: asyncio.Future | None):
    """Safely cancels a task."""

    if task is None or task.done():
        return None

    task.cancel()
    with suppress(asyncio.CancelledError):
        await task

    return None


def render_template(
    template: str | None = None,
    file: str | os.PathLike | None = None,
    render_data: dict[str, Any] = {},
):
    """Renders a template using Jinja2.

    Parameters
    ----------
    template
        The template to render. If `None`, `file` must be defined.
    file
        The path to the file to render. If `None`, `template` must be defined.
    render_data
        The data to pass to the template.

    """

    if template is not None and file is not None:
        raise ValueError("Only one of template or file can be defined.")

    if template is not None:
        templates_path = pathlib.Path(__file__).parent / "templates"
        loader = FileSystemLoader(templates_path)
    elif file is not None:
        dirname = pathlib.Path(file).parent
        loader = FileSystemLoader(dirname)
        template = dirname.name
    else:
        raise ValueError("Either template or file must be defined.")

    env = Environment(
        loader=loader,
        lstrip_blocks=True,
        trim_blocks=True,
    )
    html_template = env.get_template(template)

    return html_template.render(**render_data)


class LockExistsError(RuntimeError):
    """Raised when a lock file already exists."""

    pass


async def _monitor_lockfile(
    lockfile: pathlib.Path,
    log: SDSSLogger | None = None,
    console: Console | None = None,
    on_release_callback: Callable[..., Any] | Coroutine | None = None,
):
    """Monitors a lock file and raises an error if it is deleted."""

    while True:
        if not lockfile.exists():
            if on_release_callback:
                if asyncio.iscoroutine(on_release_callback):
                    await on_release_callback
                elif asyncio.iscoroutinefunction(on_release_callback):
                    await on_release_callback()
                else:
                    on_release_callback()

            message = f"Lock file {lockfile} has been released. Fills will be aborted."

            if console:
                console.print(f"[yellow]{message}[/]")
            if log:
                log.warning(f"{message}")

            raise RuntimeError(f"{message}")

        await asyncio.sleep(1)


@contextlib.asynccontextmanager
async def ensure_lock(
    lockfile: str | os.PathLike | pathlib.Path,
    monitor: bool = False,
    on_release_callback: Callable[..., Any] | Coroutine | None = None,
    log: SDSSLogger | None = None,
    console: Console | None = None,
):
    """Ensures a lock file is created and deleted on exit.

    Raises an exception if the lock file already exists. If ``monitor=True``,
    monitors the lock file and raises an error if it is deleted while the context
    is active. Specify a ``on_release_callback`` to call a function or
    coroutine when the lock file is released externally. ``log`` and ``console``
    can be passed to log messages to the logger and stdout respectively.

    """

    lockfile = pathlib.Path(lockfile)

    if lockfile.exists():
        raise LockExistsError(f"Lock file {lockfile} already exists.")

    lockfile.touch()

    monitor_task: asyncio.Task | None = None

    try:
        if monitor:
            monitor_task = asyncio.create_task(
                _monitor_lockfile(
                    lockfile,
                    console=console,
                    log=log,
                    on_release_callback=on_release_callback,
                )
            )
        yield
    finally:
        await cancel_task(monitor_task)
        lockfile.unlink(missing_ok=True)


def get_fake_logger():
    """Gets a logger with a disabled handler."""

    logger = getLogger(__name__)
    logger.disabled = True

    return logger


@Retrier(max_attempts=3, delay=0.5)
async def o2_alert(route: str = "http://lvm-hub.lco.cl:8090/api/alerts"):
    """Is there an active O2 alert?"""

    # HOTFIX: disable this check temporarily since one sensor is faulty.
    return False

    async with httpx.AsyncClient(follow_redirects=True) as client:
        response = await client.get(route)

    if response.status_code != 200:
        raise RuntimeError(response.text)
    else:
        alerts = response.json()
        if len(alerts["o2_room_alerts"]) == 0:
            raise RuntimeError("No O2 alert data returned by the API.")
        else:
            # This is true if any of the room alerts are active.
            return alerts["o2_alert"]


@Retrier(max_attempts=3, delay=0.5)
async def ln2_estops():
    """Returns :obj:`True` if any of the LN2 emergency stops are active."""

    async with CluClient() as client:
        try:
            status = await client.send_command("lvmecp", "status")
            safety_labels = status.replies.get("safety_status_labels").split(",")
            return "E_STOP_LN2" in safety_labels
        except Exception as ee:
            raise RuntimeError(f"Error reading estops: {ee}")


def add_json_handler(log: logging.Logger, json_path: os.PathLike | pathlib.Path):
    """Adds a JSON handler to a logger."""

    json_handler = logging.FileHandler(str(json_path), mode="w")
    json_handler.setLevel(5)
    json_handler.setFormatter(CustomJsonFormatter())
    log.addHandler(json_handler)

    return json_handler


def date_json(date: datetime | None) -> str | None:
    """Serialises a datetime object to a JSON string."""

    return date.isoformat() if date else None


class DBHandler:
    """Handles writing the fill to the database.

    Parameters
    ----------
    action
        The action being performed.
    handler
        The `.LN2Handler` instance.
    config
        The configuration object.
    api_db_route
        The API route to write the data to the database.
    json_handler
        The logging handler used to write JSON data.

    """

    def __init__(
        self,
        action: str,
        handler: LN2Handler,
        config: Config,
        api_route: str | None = None,
        json_handler: FileHandler | None = None,
    ) -> None:
        self.pk: int | None = None

        self.action = action
        self.handler = handler
        self.config = config
        self.api_route = api_route or config.internal_config["api_routes.register_fill"]

        self.json_handler = json_handler

        self.complete: bool = False
        self.plot_paths: dict[str, pathlib.Path] = {}
        self.error: Exception | str | None = None

    def get_log_data(self):
        """Returns the log data for the fill."""

        if self.json_handler:
            self.json_handler.flush()
            json_path = pathlib.Path(self.json_handler.baseFilename)
            with json_path.open("r") as ff:
                return [json.loads(line) for line in ff.readlines()]

        return None

    async def update(
        self,
        complete: bool | None = None,
        plot_paths: dict[str, pathlib.Path] | None = None,
        error: Exception | str | None = None,
        raise_on_error: bool = False,
    ):
        """Records the fill to the database.

        Parameters
        ----------
        complete
            Whether the action is complete.
        plot_paths
            A dictionary with the paths to the plots.
        error
            The error message or ``None`` if no error.
        raise_on_error
            Whether to raise an exception if the write fails.

        Returns
        -------
        pk
            The primary key of the new record in the database.

        """

        log = self.handler.log
        event_times = self.handler.event_times
        log_path = self.config.log_path

        json_path = self.json_handler.baseFilename if self.json_handler else None
        json_file = str(json_path) if json_path and self.config.write_json else None

        configuration_json = self.config.model_dump() | {
            valve: valve_model.model_dump()
            for valve, valve_model in self.config.valve_info.items()
        }

        # Use cached values
        self.complete = complete if complete is not None else self.complete
        self.plot_paths = plot_paths if plot_paths is not None else self.plot_paths
        self.error = error if error is not None else self.error

        payload = {
            "action": self.action,
            "complete": self.complete,
            "pk": self.pk,
            "start_time": date_json(event_times.start_time),
            "end_time": date_json(event_times.end_time),
            "purge_start": date_json(event_times.purge_start),
            "purge_complete": date_json(event_times.purge_complete),
            "fill_start": date_json(event_times.fill_start),
            "fill_complete": date_json(event_times.fill_complete),
            "fail_time": date_json(event_times.fail_time),
            "abort_time": date_json(event_times.abort_time),
            "failed": self.handler.failed,
            "aborted": self.handler.aborted,
            "plot_paths": {k: str(v) for k, v in self.plot_paths.items()},
            "log_file": str(log_path) if log_path else None,
            "valve_times": self.handler.get_valve_times(as_string=True),
            "json_file": json_file,
            "log_data": self.get_log_data(),
            "configuration": configuration_json,
            "error": str(self.error) if self.error is not None else None,
        }

        async with httpx.AsyncClient(follow_redirects=True) as client:
            response = await client.post(self.api_route, json=payload)

            if response.status_code != 200:
                if raise_on_error:
                    raise RuntimeError(f"Error writing to the DB: {response.text}")
                else:
                    log.warning(f"Error writing to the DB: {response.text}")
                    log.warning(f"DB payload: {payload}")

                return self.pk

            self.pk = response.json()

        return self.pk


def run_command(
    command: str | Sequence[str],
    output_on_error: bool = False,
    raise_on_error: bool = False,
    console: Console | None = None,
    **kwargs,
) -> subprocess.CompletedProcess:
    """Runs a shell command and returns the completed process.

    Parameters
    ----------
    command
        The command to run.
    output_on_error
        If :obb:`True`, outputs the error message with stdout and stderr
        if the command fails.
    raise_on_error
        If :obj:`True`, raises a :obj:`RuntimeError` if the command fails.
    console
        An optional :obj:`rich.console.Console` to use for output.
    **kwargs
        Additional arguments to pass to :obj:`subprocess.run`.

    Returns
    -------
    subprocess.CompletedProcess
        The completed process.

    Raises
    ------
    RuntimeError
        If the command fails.

    """

    console = console or Console()

    text = kwargs.pop("text", True)
    capture_output = kwargs.pop("capture_output", True)

    result = subprocess.run(
        command,
        text=text,
        capture_output=capture_output,
        **kwargs,
    )

    if result.returncode == 0:
        return result

    command = command if isinstance(command, str) else " ".join(command)
    error_msg = f"Command '{command}' failed with return code {result.returncode}."

    if output_on_error:
        console.print(f"[red]{error_msg}[/red]")
        if result.stdout:
            console.print(f"[yellow]\nSTDOUT:[/yellow]\n{result.stdout}")
        if result.stderr:
            console.print(f"[red]\nSTDERR:[/red]\n{result.stderr}")

    if raise_on_error:
        raise RuntimeError(error_msg)

    return result


def register_parameter_origin(func):
    """A decorator that registers the origin of parameters passed to a function.

    The decorated function must accept an additional keyword argument
    ``__parameter_origin`` that will be set to a mapping of parameter names to
    their :obj:`ParameterOrigin` value.

    """

    @wraps(func)
    def inner(*args, **kwargs):
        # If a __parameter_origin has already been passed, use it.
        __parameter_origin = kwargs.pop("__parameter_origin", None)
        if __parameter_origin is not None and __parameter_origin != {}:
            return func(*args, __parameter_origin=__parameter_origin, **kwargs)

        # Determine the origin of each parameter.
        names = inspect.getfullargspec(func).args
        origin: dict[str, ParameterOrigin] = {}

        for arg in names:
            if arg == "__parameter_origin":
                continue
            elif arg in kwargs:
                origin[arg] = ParameterOrigin.FUNCTION_CALL
            else:
                origin[arg] = ParameterOrigin.DEFAULT

        return func(
            *args,
            __parameter_origin=origin,
            **kwargs,
        )

    return inner


def get_lockfile_path(
    lockfile: str | os.PathLike | pathlib.Path | None = None,
) -> pathlib.Path:
    """Returns the path to the lock file."""

    lockfile = lockfile or get_internal_config().get("lockfile", "/data/lvmcryo.lock")

    return pathlib.Path(lockfile)


def lockfile_exists(lockfile: str | os.PathLike | pathlib.Path | None = None) -> bool:
    """Checks if the lock file exists."""

    return get_lockfile_path(lockfile).exists()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-12-03
# @Filename: runner.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import inspect
import itertools
import logging
import os
import pathlib
import signal
import sys
from datetime import timedelta
from logging import FileHandler
from tempfile import NamedTemporaryFile

from typing import TYPE_CHECKING, Literal

import httpx
import polars
from rich.prompt import Confirm

from sdsstools.logger import get_logger
from sdsstools.utils import run_in_executor

from lvmcryo import __version__
from lvmcryo.config import (
    Actions,
    Config,
    InteractiveMode,
    NotificationLevel,
    ParameterOrigin,
    get_internal_config,
)
from lvmcryo.handlers import LN2Handler, close_all_valves
from lvmcryo.handlers.ln2 import get_now
from lvmcryo.notifier import Notifier
from lvmcryo.tools import (
    DBHandler,
    LockExistsError,
    add_json_handler,
    ensure_lock,
    get_lockfile_path,
    register_parameter_origin,
)
from lvmcryo.validate import validate_fill


if TYPE_CHECKING:
    from sdsstools.logger import SDSSLogger

__all__ = ["fill_runner", "ln2_runner"]


async def signal_handler(handler: LN2Handler, log: logging.Logger):
    """Handles signals to close all valves and exit cleanly."""

    handler.failed = True
    handler.aborted = True

    log.error("User aborted the process. Closing all valves before exiting.")
    try:
        await handler.stop(only_active=False)
    except Exception as err:
        log.error(f"Failed closing LN2 valves: {err!r}")

    await handler.clear()

    handler.event_times.fail_time = get_now()
    handler.event_times.abort_time = get_now()
    handler.event_times.end_time = get_now()

    log.error("Exiting now. No data or notifications will be sent.")

    # Clear the lock.
    get_lockfile_path().unlink(missing_ok=True)

    # Prevent the event loop from throwing some exception when we exit.
    asyncio.get_running_loop().stop()

    sys.exit(1)


class LN2RunnerError(Exception):
    """An error occurred during the LN2 runner execution."""

    def __init__(self, *args, propagate: bool = False, **kwargs):
        super().__init__(*args, **kwargs)
        self.propagate = propagate


@register_parameter_origin
async def ln2_runner(
    action: Literal["purge", "fill", "purge-and-fill"] | Actions = "purge-and-fill",
    cameras: list[str] | None = None,
    profile: str | None = None,
    config_file: pathlib.Path | str | None = None,
    dry_run: bool = False,
    interactive: Literal["auto", True, False, "yes", "no"] | InteractiveMode = False,
    no_prompt: bool = True,
    with_traceback: bool = False,
    clear_lock: bool = False,
    use_thermistors: bool = True,
    require_all_thermistors: bool = False,
    check_pressures: bool = True,
    check_temperatures: bool = True,
    check_o2_sensors: bool = True,
    max_pressure: float | None = None,
    max_temperature: float | None = None,
    purge_time: float | None = None,
    min_purge_time: float | None = None,
    max_purge_time: float | None = None,
    fill_time: float | None = None,
    min_fill_time: float | None = None,
    max_fill_time: float | None = None,
    notify: bool = False,
    slack: bool = True,
    email: bool = True,
    email_level: Literal["error", "info"] | NotificationLevel = "error",
    quiet: bool = False,
    verbose: bool = False,
    write_log: bool = False,
    log_path: pathlib.Path | None = None,
    write_json: bool = True,
    write_data: bool = False,
    data_path: pathlib.Path | None = None,
    data_extra_time: float = 60,
    log: SDSSLogger | None = None,
    __parameter_origin: dict[str, ParameterOrigin | None] = {},
):
    """Runs LN2 purge/fill/abort/clear actions.

    This is a wrapper allowing the CLI functionality to be called programmatically.
    Note that the default values in this function are only for reference; the actual
    default values are defined in the ``defaults`` section of the configuration file
    or as the default field value in the :obj:`.Config` model.

    Parameters are processed in this order:
    - If passed as function arguments, those values are used.
    - If not passed but an environment variable ``LVMCRYO_<PARAMETER_NAME>`` is set,
      that value is used.
    - If a profile is specified, parameters defined in that profile are used.
      The profile parameters are overridden if the parameter is specified as
      in the function call or with an environment variable.
    - Otherwise the default value in the configuration file or the model is used.

    Parameters
    ----------
    action
        The action to perform. One of ``purge``, ``fill``, or ``purge-and-fill``.
    cameras
        The cameras to fill. Defaults to all cameras. Ignored for ``purge``, ``abort``
        and ``clear_lock``.
    profile
        The profile to use.
    config_file
        The configuration file. Defaults to the internal configuration file.
    dry_run
        Test run the code but do not actually open/close any valves.
    interactive
        Controls whether the interactive features are shown.
        When ``auto`` (the default), interactivity is determined based
        on the console mode.
    no_prompt
        Does not prompt the user to finish or abort a purge/fill.
        Prompting is required if ``fill_time`` or ``purge_time`` are not provided
        and thermistors are not used.
    with_traceback
        Show the full traceback in case of an error. If not set, only the error message
        is shown. The full traceback is always logged to the log file.
    clear_lock
        Clears the lock file if it exists before carrying out the action.
    use_thermistors
        Use thermistor values to determine purge/fill time.
    require_all_thermistors
        If set, waits until all thermistors have activated before closing any of the
        valves. This prevents over-pressures in the cryostats when only some
        of the valves are open. Ignored if ``use_thermistors=False``.
    check_pressures
        Aborts purge/fill if the pressure of any cryostat is above the limit.
    check_temperatures
        Aborts purge/fill if the temperature of a fill cryostat is above the limit.
    check_o2_sensors
        Aborts purge/fill if the oxygen sensor reads below the limit.
    max_pressure
        Maximum cryostat pressure if ``--check-pressure``. Defaults to internal value.
    max_temperature
        Maximum cryostat temperature if ``check_temperature=True``.
        Defaults to internal value.
    purge_time
        Purge time in seconds. If ``use_thermistors=True``, this is used
        as the maximum purge time.
    min_purge_time
        Minimum purge time in seconds. Defaults to internal value.
    max_purge_time
        Maximum purge time in seconds. Defaults to internal value.
    fill_time
        Fill time in seconds. If ``use_thermistors=True``, this is used as
        the maximum fill time.
    min_fill_time
        Minimum fill time in seconds. Defaults to internal value.
    max_fill_time
        Maximum fill time in seconds. Defaults to internal value.
    notify
        Sends a notification when the action is completed. In not set,
        ``slack`` and ``email`` are ignored. Notifications are not sent
        for the ``abort`` and ``clear-lock`` actions.
    slack
        Send a Slack notification when the action is completed.
    email
        Send an email notification when the action is completed.
    email_level
        Send an email notification only on error or always.
    quiet
        Disable logging to stdout.
    verbose
        Outputs additional information to stdout.
    write_log
        Saves the log to a file.
    log_path
        Path where to save the log file. Implies ``write_log=True``.
        Defaults to the current directory.
    write_json
        Saves the log in JSON format. Uses the same path as the file log. Ignored if
        ``write_log`` is not passed.
    write_data
        Saves cryostat pressures and temperatures taken during purge/fill to a
        parquet file.
    data_path
        Path where to save the data. Implies ``write_data=True``.
        Defaults to a path relative to the log path.
    data_extra_time
        Additional time to take cryostat data after the action has been completed.
        The command will not complete until data collection has finished.
    log
        The logger instance to use. If `None`, a new logger is created.
    __parameter_origin
        Used internally to track parameter origins. Do not pass this parameter
        manually.

    """

    try:
        action = Actions(action)

        if interactive is True:
            interactive = "yes"
        elif interactive is False:
            interactive = "no"
        interactive = InteractiveMode(interactive)

        email_level = NotificationLevel(email_level)

        # Get profile and config_file from envvars
        # if we have not explicitly passed them.
        profile = profile or os.environ.get("LVMCRYO_PROFILE", None)

        config_file = config_file or os.environ.get("LVMCRYO_CONFIG_FILE", None)
        config_file = pathlib.Path(config_file) if config_file else None

        # Determine which parameters were provided on the CLI or as function
        # arguments. For parameters not passed, check if there is an environment
        # variable set. If so, use that value.
        config_params = {}

        args = list(inspect.signature(ln2_runner).parameters.keys())
        for arg in args:
            if arg in ["__parameter_origin", "log"]:
                continue

            param_value = locals()[arg]
            param_origin = __parameter_origin.get(arg, ParameterOrigin.DEFAULT)

            if param_origin == ParameterOrigin.DEFAULT:
                # Check if the environment variable is set. Otherwise skip and let
                # the default value be set in the Config class.
                envvar = f"LVMCRYO_{arg.upper()}"
                envvar_value = os.environ.get(envvar, None)
                if envvar_value is not None:
                    config_params[arg] = envvar_value
            elif param_origin == ParameterOrigin.ENVVAR:
                config_params[arg] = param_value
            elif param_origin == ParameterOrigin.FUNCTION_CALL:
                config_params[arg] = param_value
            elif param_origin == ParameterOrigin.COMMAND_LINE:
                config_params[arg] = param_value

        # Clear the cache for get_internal_config() to ensure fresh load.
        get_internal_config.cache_clear()

        # Get the internal config and defaults.
        config = get_internal_config(config_file)

        # If we have specified a profile, use those parameters except if we have
        # overridden them with a function parameters or envvar.
        if profile is not None:
            profile_config: dict = config[f"profiles.{profile}"] or {}
            for key in profile_config:
                if key not in config_params:
                    config_params[key] = profile_config[key]

        config = Config(**config_params, version=__version__)
        internal_config = config.internal_config
    except Exception as err:
        raise LN2RunnerError(f"Error parsing configuration: {err!r}") from err

    # Create log here and use its console for stdout.
    if log is None:
        log = get_logger("lvmcryo", use_rich_handler=True)
        log.setLevel(5)

    stdout_console = log.rich_console
    assert stdout_console is not None

    error: Exception | None = None
    skip_finally: bool = False

    json_path: pathlib.Path | None = None
    json_handler: FileHandler | None = None
    images: dict[str, pathlib.Path | None] = {}

    if config.write_log and config.log_path:
        log.start_file_logger(str(config.log_path))

        if config.write_json:
            json_path = config.log_path.with_suffix(".json")
            json_handler = add_json_handler(log, json_path)

    else:
        # We're still creating log files, but to a temporary location. This is
        # just to be able to send the log body in a notification email and include
        # it when loading the DB.
        temp_file = NamedTemporaryFile()
        log.start_file_logger(temp_file.name)
        json_path = pathlib.Path(temp_file.name).with_suffix(".json")
        json_handler = add_json_handler(log, json_path)

    if verbose:
        log.sh.setLevel(5)
    if quiet:
        log.sh.setLevel(30)

    # Always create a notifier. If the element is disabled the
    # notifier won't do anything. This simplifies the code.
    notifier = Notifier(internal_config)
    notifier.disabled = not config.notify
    notifier.slack_disabled = not config.slack
    notifier.email_disabled = not config.email

    if not config.notify:
        log.debug("Notifications are disabled and will not be emitted.")

    if config.config_file is not None:
        log.info(f"Using configuration file: {config.config_file!s}")

    if not config.no_prompt:
        stdout_console.print(f"Action {config.action.value} will run with:")
        stdout_console.print(config.model_dump())
        if not Confirm.ask(
            "Continue with this configuration?",
            default=False,
            show_default=True,
            console=stdout_console,
        ):
            return False

    # Create the LN2 handler.
    try:
        handler = LN2Handler(
            config.cameras,
            interactive=config.interactive == "yes",
            log=log,
            valve_info=config.valve_info,
            dry_run=config.dry_run,
            alerts_route=internal_config["api_routes"]["alerts"],
            check_o2_sensors=config.check_o2_sensors,
        )
    except Exception as err:
        raise LN2RunnerError(
            f"Error creating LN2 handler: {err!r}",
            propagate=config.with_traceback,
        ) from err

    lockfile_path = get_lockfile_path()

    if lockfile_path.exists() and config.clear_lock:
        log.warning("Lock file exists. Removing it because --clear-lock.")
        lockfile_path.unlink()

        log.info("Waiting 10 seconds for other processes to stop.")
        await asyncio.sleep(10)

    try:
        db_handler = DBHandler(action, handler, config, json_handler=json_handler)
        record_pk = await db_handler.update(complete=False)
        if record_pk:
            log.debug(f"Record {record_pk} created in the database.")
    except Exception as err:
        raise LN2RunnerError(
            f"Error creating database record: {err!r}",
            propagate=config.with_traceback,
        ) from err

    try:
        async with ensure_lock(
            lockfile_path,
            monitor=True,
            log=log,
            on_release_callback=handler.abort(raise_error=False, close_valves=True),
        ):
            # Calculate the expected maximum run time.
            max_time: float = 2 * 3600  # It should never take longer than two hours.
            if config.max_purge_time is not None and config.max_fill_time is not None:
                max_time = config.max_purge_time + config.max_fill_time + 300.0

            # Run worker.
            await asyncio.wait_for(
                fill_runner(
                    handler,
                    config,
                    notifier,
                    db_handler=db_handler,
                ),
                timeout=max_time,
            )

            # Check handler status.
            if handler.failed:
                raise RuntimeError(
                    "No exceptions were raised but the LN2 handler reports a failure."
                )

    except LockExistsError:
        if config.notify:
            log.warning("Sending failure notifications.")
            await notifier.notify_after_fill(
                False,
                error_message=f"LN2 {config.action.value} failed because a lockfile "
                "was already present.",
            )

        # Do not do anything special for this error, just exit.
        skip_finally = True

        raise

    except Exception as err:
        # Log the traceback to file but do not print.
        orig_sh_level = log.sh.level
        log.sh.setLevel(1000)

        log.exception(f"Error during {config.action.value}: {err!s}", exc_info=err)
        if isinstance(err, asyncio.TimeoutError):
            log.error("One or more operations timed out.")

        log.sh.setLevel(orig_sh_level)

        # Fail the action.
        handler.failed = True
        skip_finally = True

        error = err
        raise LN2RunnerError(str(err), propagate=config.with_traceback) from err

    else:
        log.info(f"LN2 {config.action.value} completed successfully.")

    finally:
        # At this point all the valves are closed so we can remove the signal handlers.
        for signame in ("SIGINT", "SIGTERM"):
            asyncio.get_running_loop().remove_signal_handler(getattr(signal, signame))

        handler.event_times.end_time = get_now()
        await handler.clear()

        log.info(f"Event times:\n{handler.event_times.model_dump_json(indent=2)}")

        # Make sure all valves are closed.
        try:
            log.info("Ensuring all valves are closed.")
            await asyncio.wait_for(
                handler.stop(only_active=False, close_valves=True),
                timeout=30,
            )
        except Exception as err:
            log.error(f"Error closing valves before exiting: {err}")

        # Do a quick update of the DB record since post_fill_tasks() may
        # block for a long time.
        await db_handler.update(complete=True, error=error)

        if not skip_finally:
            plot_paths = await post_fill_tasks(
                handler,
                notifier=notifier,
                write_data=config.write_data,
                data_path=config.data_path,
                data_extra_time=config.data_extra_time if error is None else None,
                api_data_route=internal_config["api_routes"]["fill_data"],
            )

            if (
                config.write_data
                and config.data_path
                and config.data_path.exists()
                and not error
            ):
                validate_failed, validate_error = validate_fill(
                    handler,
                    config,
                    log=log,
                )
                if validate_failed and error is None:
                    await notifier.post_to_slack(
                        "Fill validation failed. Check the log for details.",
                        level=NotificationLevel.error,
                    )
                    handler.failed = True
                    error = RuntimeError(validate_error)
                elif not validate_failed:
                    log.info("Fill validation completed successfully.")

            log.info("Writing fill metadata to database.")
            await db_handler.update(complete=True, plot_paths=plot_paths, error=error)

            if config.notify:
                images = {
                    "pressure": plot_paths.get("pressure_png", None),
                    "temps": plot_paths.get("temps_png", None),
                    "thermistors": plot_paths.get("thermistors_png", None),
                }

                if error:
                    log.warning("Sending failure notifications.")
                    await notifier.notify_after_fill(
                        False,
                        error_message=error,
                        handler=handler,
                        images=images,
                        record_pk=record_pk,
                    )

                elif config.email_level == NotificationLevel.info:
                    # The handler has already emitted a notification to
                    # Slack so just send an email.

                    # TODO: include log and more data here.
                    # For now it's just plain text.

                    log.info("Sending notification email.")
                    await notifier.notify_after_fill(
                        True,
                        handler=handler,
                        images=images,
                        post_to_slack=False,  # Already done.
                        record_pk=record_pk,
                    )

        await db_handler.update()


async def fill_runner(
    handler: LN2Handler,
    config: Config,
    notifier: Notifier | None = None,
    db_handler: DBHandler | None = None,
):
    """Runs the purge/fill process.

    This function is usually called by the CLI.

    """

    if notifier is None:
        # Create a notifier but immediately disable it.
        notifier = Notifier()
        notifier.disabled = True

    log = handler.log

    # Record options used. If in no_promp is False (the default), the full
    # configuration has already been printed for confirmation, so we only
    # save this with debug level.
    config_json = config.model_dump_json(indent=2)
    log.log(
        logging.INFO if config.no_prompt else logging.DEBUG,
        f"Running {config.action.value} with configuration:\n{config_json}",
    )

    if config.dry_run:
        log.warning("Running in dry-run mode. No valves will be operated.")

    # Inform of the start of the fill in Slack.
    now_str = get_now().strftime("%H:%M:%S")

    action = config.action.value

    if config.use_thermistors or config.purge_time is None or config.fill_time is None:
        lvmweb_url = config.internal_config["notifications.lvmweb_fill_url"]
        message = f"Starting LN₂ `{action}` at {now_str}."
        if db_handler and db_handler.pk is not None:
            fill_url = lvmweb_url.format(fill_id=db_handler.pk)
            message += f" Details can be followed from the <{fill_url}|LVM webapp>."

        await notifier.post_to_slack(message)

    else:
        await notifier.post_to_slack(
            f"Starting LN₂ `{action}` at {now_str} with "
            f"purge_time={config.purge_time} and "
            f"fill_time={config.fill_time}."
        )

    max_temperature = config.max_temperature if config.check_temperatures else None
    max_pressure = config.max_pressure if config.check_pressures else None
    await handler.check(
        max_pressure=max_pressure,
        max_temperature=max_temperature,
        check_thermistors=config.use_thermistors,
        check_o2_sensors=config.check_o2_sensors,
    )

    # Register signals that will trigger a valve shutdown and clean exit.
    for signame in ("SIGINT", "SIGTERM"):
        asyncio.get_running_loop().add_signal_handler(
            getattr(signal, signame),
            lambda: asyncio.create_task(signal_handler(handler, log)),
        )

    log.info(f"Closing all valves before {action}.")
    await close_all_valves(dry_run=config.dry_run)

    if handler.failed or handler.aborted:
        await handler.clear()
        raise RuntimeError("The LN2 handler was aborted before starting the fill.")

    if config.action == Actions.purge_fill or config.action == Actions.purge:
        await notifier.post_to_slack("Starting purge.")
        max_purge_time = config.purge_time or config.max_purge_time
        await asyncio.wait_for(
            handler.purge(
                use_thermistor=config.use_thermistors,
                min_purge_time=config.min_purge_time,
                max_purge_time=max_purge_time,
                prompt=not config.no_prompt,
                preopen_cb=db_handler.update if db_handler is not None else None,
            ),
            timeout=max_purge_time + 60 if max_purge_time else None,
        )

        if handler.failed or handler.aborted:
            await handler.clear()
            raise RuntimeError(handler.error or "Purge failed or was aborted.")

    if config.action == Actions.purge_fill or config.action == Actions.fill:
        await notifier.post_to_slack("Starting fill.")
        max_fill_time = config.fill_time or config.max_fill_time
        await asyncio.wait_for(
            handler.fill(
                use_thermistors=config.use_thermistors,
                require_all_thermistors=config.require_all_thermistors,
                min_fill_time=config.min_fill_time,
                max_fill_time=max_fill_time,
                prompt=not config.no_prompt,
                preopen_cb=db_handler.update if db_handler is not None else None,
            ),
            timeout=max_fill_time + 60 if max_fill_time else None,
        )

        if handler.failed or handler.aborted:
            await handler.clear()
            raise RuntimeError(handler.error or "Fill failed or was aborted.")

    await notifier.post_to_slack(f"LN₂ `{action}` completed successfully.")
    await handler.clear()

    handler.event_times.end_time = get_now()


async def post_fill_tasks(
    handler: LN2Handler,
    notifier: Notifier | None = None,
    write_data: bool = False,
    data_path: str | pathlib.Path | None = None,
    data_extra_time: float | None = None,
    api_data_route: str = "http://lvm-hub.lco.cl:8090/api/spectrographs/fills/measurements",
    generate_data_plots: bool = True,
) -> dict[str, pathlib.Path]:
    """Runs the post-fill tasks.

    Parameters
    ----------
    handler
        The `.LN2Handler` instance.
    write_data
        Whether to collect fill metrology data and write it to disk.
    data_path
        The path where to write the data. If `None` writes it to the current directory.
    data_extra_time
        Extra time to wait after the fill before collecting data.
    api_data_route
        The API route to retrive the fill data.
    generate_data_plots
        Whether to generate plots from the data.

    Returns
    -------
    plot_paths
        A mapping of plot type to path.

    """

    log = handler.log
    log.info("Running post-fill tasks.")

    event_times = handler.event_times
    plot_paths: dict[str, pathlib.Path] = {}

    if (
        not write_data
        or not event_times.start_time
        or not event_times.end_time
        or not api_data_route
    ):
        write_data = False
        log.debug("Skipping data collection.")

    if event_times.start_time and event_times.end_time:
        interval = (event_times.end_time - event_times.start_time).total_seconds()
        if interval < 120:
            log.warning("Fill duration was less than 2 minutes. Not collecting data.")
            write_data = False

    if write_data and event_times.start_time and event_times.end_time:
        if data_extra_time:
            if handler.aborted or handler.failed:
                log.warning(
                    "Fill was aborted or failed. Not waiting "
                    "additional time to collect data."
                )
            else:
                log.info(f"Waiting {data_extra_time} seconds before collecting data.")

                if notifier is not None:
                    await notifier.post_to_slack(
                        f"Fill notifications will be delayed {data_extra_time:.0f} "
                        "seconds while collecting post-fill data."
                    )

                await asyncio.sleep(data_extra_time)

        if data_path is None:
            data_path = pathlib.Path.cwd() / "fill_data.parquet"
        else:
            data_path = pathlib.Path(data_path)

        try:
            log.info("Retrieving and writing measurements.")

            end_time = event_times.end_time + timedelta(seconds=data_extra_time or 0.0)
            async with httpx.AsyncClient(follow_redirects=True) as client:
                response = await client.get(
                    api_data_route,
                    params={
                        "start_time": int(event_times.start_time.timestamp()),
                        "end_time": int(end_time.timestamp()),
                    },
                )
                response.raise_for_status()

                data = (
                    polars.DataFrame(response.json())
                    .with_columns(polars.col.time.cast(polars.Datetime("ms")))
                    .sort("time")
                    .drop_nulls()
                )

                data_path.parent.mkdir(parents=True, exist_ok=True)
                data.write_parquet(data_path)
                log.debug(f"Fill data written to {data_path!s}")

                if generate_data_plots:
                    log.debug("Generating plots.")
                    plot_path_root = str(data_path.with_suffix(""))
                    plot_paths = await run_in_executor(
                        generate_plots,
                        data,
                        plot_path_root,
                    )
                    plot_paths_transparent = await run_in_executor(
                        generate_plots,
                        data,
                        plot_path_root,
                        transparent=True,
                    )
                    plot_paths.update(plot_paths_transparent)
                    log.debug(f"Plots saved to {plot_path_root}*.")

        except Exception as ee:
            log.error(f"Failed to retrieve fill data from API: {ee!r}")

    return plot_paths


def generate_plots(
    data: polars.DataFrame,
    plot_path_root: str,
    transparent: bool = False,
    include_ccd_tempratures: bool = False,
):
    """Generates measurement plots.

    Parameters
    ----------
    data
        The data to plot.
    plot_path_root
        The root path where to save the plots. A number of plots will be generated
        using this root path.
    transparent
        Whether to save the plots with a transparent background. The colours will
        be adjusted to be visible on a dark background. The paths will be suffixed
        with ``_transparent``. No PDFs will be generated in this case.
    include_ccd_tempratures
        Whether to include CCD temperatures in the temperature plot.

    Returns
    -------
    paths
        A mapping of plot type to path.

    """

    import matplotlib.pyplot as plt

    paths: dict[str, pathlib.Path] = {}

    colours: dict[str, str] = {
        "r": "red",
        "b": "cyan" if transparent else "blue",
        "z": "magenta",
    }
    linestyles: dict[str, str] = {"1": "-", "2": "--", "3": "-."}

    transparent_suffix = "_transparent" if transparent else ""

    date = data[0, "time"].strftime("%Y-%m-%d")

    with plt.ioff():
        if transparent:
            plt.style.use("dark_background")
        else:
            plt.style.use("seaborn-v0_8-whitegrid")
            plt.rcParams["axes.facecolor"] = "white"
            plt.rcParams["figure.facecolor"] = "white"
            plt.rcParams["savefig.facecolor"] = "white"

        # Pressures
        fig, ax = plt.subplots(figsize=(12, 8))

        pressures = data.select(
            polars.col.time,
            polars.selectors.starts_with("pressure_"),
        )

        for spec, camera in itertools.product("123", "brz"):
            column = f"pressure_{camera}{spec}"
            if column not in pressures.columns:
                continue

            cam_pressure = pressures[column].to_numpy()

            colour = colours.get(camera, "w" if transparent else "k")
            linestyle = linestyles.get(spec, "-")

            label = f"{camera}{spec}"

            ax.plot(
                pressures["time"].to_numpy(),
                cam_pressure,
                label=label,
                color=colour,
                linestyle=linestyle,
            )

        ax.set_title(f"Pressure during fill — {date}")
        ax.set_xlabel("Time")
        ax.set_ylabel("Pressure [torr]")

        ax.legend(loc="center left", bbox_to_anchor=(1, 0.5), prop={"size": 9})

        plt.ticklabel_format(style="sci", axis="y", scilimits=(0, 0))

        if not transparent:
            path = f"{plot_path_root}_pressure{transparent_suffix}.pdf"
            fig.savefig(path)
            paths[f"pressure{transparent_suffix}_pdf"] = pathlib.Path(path)

        path = f"{plot_path_root}_pressure{transparent_suffix}.png"
        fig.savefig(path, dpi=300, transparent=transparent)
        paths[f"pressure{transparent_suffix}_png"] = pathlib.Path(path)

        plt.close(fig)

        # Temperatures
        fig, ax = plt.subplots(figsize=(12, 8))

        temps = data.select(
            polars.col.time,
            polars.selectors.starts_with("temp_"),
        )

        for spec, camera, sensor in itertools.product("123", "brz", ["ln2", "ccd"]):
            if sensor == "ccd" and not include_ccd_tempratures:
                continue

            column = f"temp_{camera}{spec}_{sensor}"
            if column not in temps.columns:
                continue

            cam_temp = temps[column].to_numpy()

            colour = colours.get(camera, "w" if transparent else "k")
            linestyle = linestyles.get(spec, "-")
            linewidth = 1.5 if sensor == "ln2" else 1

            label = f"{camera}{spec} ({sensor.upper()})"

            ax.plot(
                temps["time"].to_numpy(),
                cam_temp,
                label=label,
                color=colour,
                linestyle=linestyle,
                linewidth=linewidth,
            )

        ax.set_title(f"Temperature during fill — {date}")
        ax.set_xlabel("Time")
        ax.set_ylabel("Temperature [C]")

        ax.legend(loc="center left", bbox_to_anchor=(1, 0.5), prop={"size": 9})

        if not transparent:
            path = f"{plot_path_root}_temps{transparent_suffix}.pdf"
            fig.savefig(path)
            paths[f"temps{transparent_suffix}_pdf"] = pathlib.Path(path)

        path = f"{plot_path_root}_temps{transparent_suffix}.png"
        fig.savefig(path, dpi=300, transparent=transparent)
        paths[f"temps{transparent_suffix}_png"] = pathlib.Path(path)

        plt.close(fig)

        # Thermistors
        fig, ax = plt.subplots(figsize=(12, 8))

        therms = data.select(
            polars.col.time,
            polars.selectors.starts_with("thermistor_"),
        )

        cameras = ["".join(item)[::-1] for item in itertools.product("123", "brz")]
        for channel in ["supply"] + cameras:
            column = f"thermistor_{channel}"
            if column not in therms.columns:
                continue

            cam_therm = therms[column].to_numpy()

            if len(channel) == 2:
                camera, spec = channel
                colour = colours.get(camera, "w" if transparent else "k")
                linestyle = linestyles.get(spec, "-")
            else:
                colour = "g" if transparent else "k"
                linestyle = "-"

            ax.plot(
                therms["time"].to_numpy(),
                cam_therm,
                label=channel,
                color=colour,
                linestyle=linestyle,
            )

        ax.set_title(f"Thermistors during fill — {date}")
        ax.set_xlabel("Time")
        ax.set_ylabel("State")

        ax.legend(loc="center left", bbox_to_anchor=(1, 0.5), prop={"size": 9})

        if not transparent:
            path = f"{plot_path_root}_thermistors{transparent_suffix}.pdf"
            fig.savefig(path)
            paths[f"thermistors{transparent_suffix}_pdf"] = pathlib.Path(path)

        path = f"{plot_path_root}_thermistors{transparent_suffix}.png"
        fig.savefig(path, dpi=300, transparent=transparent)
        paths[f"thermistors{transparent_suffix}_png"] = pathlib.Path(path)

        plt.close(fig)

    return paths


async def clear_lock(
    lockfile_path: str | pathlib.Path | None = None,
    wait: bool = True,
    wait_delay: float = 10,
    log: SDSSLogger | None = None,
):
    """Clears the lock file if it exists.

    Parameters
    ----------
    lockfile_path
        The path to the lock file. If :obj:`None`, the internal configuration
        value will be used.
    wait
        Whether to wait for other processes to stop after removing the lock file.
    wait_delay
        Time in seconds to wait after removing the lock file.
    log
        The logger instance to use. If not passed the operation is silent.

    """

    lockfile_path = get_lockfile_path(lockfile_path)

    if lockfile_path.exists():
        if log:
            log.warning("Lock file exists. Removing it.")
        lockfile_path.unlink()

        if wait:
            if log:
                log.info(f"Waiting {wait_delay} seconds for other processes to stop.")
                await asyncio.sleep(wait_delay)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-09-14
# @Filename: config.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import datetime
import enum
import os
import pathlib
import warnings
from enum import Enum
from functools import lru_cache

from typing import Annotated, Any, Self

from pydantic import (
    BaseModel,
    Field,
    PrivateAttr,
    field_serializer,
    field_validator,
    model_validator,
)

from sdsstools.configuration import Configuration


class Actions(str, Enum):
    purge_fill = "purge-and-fill"
    purge = "purge"
    fill = "fill"


class InteractiveMode(str, Enum):
    auto = "auto"
    yes = "yes"
    no = "no"


class NotificationLevel(str, Enum):
    info = "info"
    error = "error"


class ParameterOrigin(Enum):
    """Enumeration to track how a parameter was set."""

    DEFAULT = enum.auto()
    ENVVAR = enum.auto()
    COMMAND_LINE = enum.auto()
    FUNCTION_CALL = enum.auto()


class ThermistorConfig(BaseModel):
    """Thermistor configuration model."""

    channel: str | None = Field(default=None)
    monitoring_interval: float = Field(default=1.0)
    close_valve: bool = Field(default=True)
    required_active_time: float = Field(default=10.0)
    disabled: bool = Field(default=False)


class ValveConfig(BaseModel):
    """Valve configuration model."""

    actor: str
    outlet: str
    thermistor: ThermistorConfig | None = Field(default_factory=ThermistorConfig)

    @field_validator("thermistor", mode="before")
    @classmethod
    def validate_thermistor(cls, value: Any) -> ThermistorConfig:
        """Validates the thermistor configuration."""

        if value is None:
            return ThermistorConfig()

        if isinstance(value, ThermistorConfig):
            return value

        return ThermistorConfig(**value)

    @model_validator(mode="after")
    def validate_after(self) -> Self:
        """Validates the valve configuration after the fields have been set."""

        if self.thermistor and self.thermistor.channel is None:
            self.thermistor.channel = self.outlet

        return self


ExcludedField = Field(repr=False, exclude=True)


@lru_cache()
def get_internal_config(path: pathlib.Path | str | None = None) -> Configuration:
    """Returns the internal configuration."""

    default_path = pathlib.Path(__file__).parent / "config.yaml"
    envvar_config_file = os.environ.get("LVMCRYO_CONFIG_FILE", None)

    if path is None and envvar_config_file is not None:
        path = envvar_config_file

    return Configuration(path, base_config=default_path)


class Config(BaseModel):
    """Configuration model.

    This model defines the options with which a purge/fill process will run.
    It is usually generated from the options passed to a CLI command after
    they are parsed by `.process_cli_options`.

    """

    # Configuration options
    profile: str | None = Field(
        default=None,
        description="The configuration profile to use.",
    )
    config_file: pathlib.Path | None = Field(
        default=None,
        description="The configuration file to use.",
    )

    # Actions
    action: Actions = Field(
        default=Actions.purge_fill,
        description="The action to perform.",
    )
    cameras: list[str] = Field(
        default_factory=list,
        description="List of camera names to use.",
    )

    # Interactivity options
    interactive: InteractiveMode = Field(
        default=InteractiveMode.auto,
        description="The interactive mode to use.",
    )
    no_prompt: bool = Field(
        default=False,
        description="Whether to disable prompts.",
    )
    dry_run: bool = Field(
        default=False,
        description="Whether to perform a dry run without opening/closing valves.",
    )
    clear_lock: bool = Field(
        default=False,
        description="Whether to clear the lock file before running.",
    )
    with_traceback: bool = Field(
        default=False,
        description="Whether to show full traceback on errors.",
    )

    # Thermistor and sensor options
    use_thermistors: bool = Field(
        default=True,
        description="Whether to use thermistors.",
    )
    require_all_thermistors: bool = Field(
        default=False,
        description="Whether to require all thermistors to be active.",
    )

    # Initial checks
    check_pressures: bool = Field(
        default=True,
        description="Whether to check pressures.",
    )
    check_temperatures: bool = Field(
        default=True,
        description="Whether to check temperatures.",
    )
    check_o2_sensors: bool = Field(
        default=True,
        description="Whether to check oxygen sensors.",
    )
    max_pressure: float | None = Field(
        default=None,
        description="Maximum cryostat pressure if checking pressures.",
    )
    max_temperature: float | None = Field(
        default=None,
        description="Maximum cryostat temperature if checking temperatures.",
    )

    # Timeout options
    purge_time: float | None = Field(
        default=None,
        description="Purge time in seconds.",
    )
    min_purge_time: float | None = Field(
        default=None,
        description="Minimum purge time in seconds.",
    )
    max_purge_time: float | None = Field(
        default=None,
        description="Maximum purge time in seconds.",
    )
    fill_time: float | None = Field(
        default=None,
        description="Fill time in seconds.",
    )
    min_fill_time: float | None = Field(
        default=None,
        description="Minimum fill time in seconds.",
    )
    max_fill_time: float | None = Field(
        default=None,
        description="Maximum fill time in seconds.",
    )
    max_temperature_increase: float = Field(
        default=0.0,
        description="Maximum temperature increase allowed.",
    )

    # Output and notification options
    verbose: bool = Field(
        default=False,
        description="Whether to enable verbose output.",
    )
    quiet: bool = Field(
        default=False,
        description="Whether to suppress output.",
    )
    notify: bool = Field(
        default=True,
        description="Whether to send notifications.",
    )
    slack: bool = Field(
        default=True,
        description="Whether to send Slack notifications.",
    )
    email: bool = Field(
        default=True,
        description="Whether to send email notifications.",
    )
    email_level: NotificationLevel = Field(
        default=NotificationLevel.error,
        description="The notification level for emails.",
    )

    # Logging options
    write_log: bool = Field(
        default=False,
        description="Whether to write logs to a file.",
    )
    log_path: pathlib.Path | None = Field(
        default=None,
        description="Path to the log file.",
    )
    write_json: bool = Field(
        default=False,
        description="Whether to write JSON output.",
    )
    write_data: bool = Field(
        default=False,
        description="Whether to write data output.",
    )
    data_path: pathlib.Path | None = Field(
        default=None,
        description="Path to the data file.",
    )
    data_extra_time: float = Field(
        default=0.0,
        description="Extra time to add to data collection.",
    )

    version: str | None = Field(
        default=None,
        description="The version of the lvmcryo package.",
    )

    valve_info: Annotated[dict[str, ValveConfig], ExcludedField] = {}

    _internal_config: Annotated[Configuration, PrivateAttr] = Configuration({})

    def model_post_init(self, __context: Any) -> None:
        self._internal_config = get_internal_config(self.config_file)
        return super().model_post_init(__context)

    @property
    def internal_config(self):
        return self._internal_config

    @field_validator("interactive", mode="after")
    @classmethod
    def validate_interactive(cls, value: InteractiveMode) -> InteractiveMode:
        """Returns the interactive mode."""

        from lvmcryo.tools import is_container

        if value in ["yes", True, 1]:
            if is_container():
                warnings.warn(
                    "Interactive mode may not work in containers.",
                    UserWarning,
                )
            return InteractiveMode.yes

        if value in ["no", False, 0]:
            return InteractiveMode.no

        if value in ["auto", None]:
            if is_container():
                return InteractiveMode.no
            else:
                return InteractiveMode.yes

        raise ValueError(f"Invalid value for interactive mode: {value!r}")

    @field_serializer("action", "interactive", "email_level")
    def serialize_enums(self, enum) -> str:
        """Serialises the enumerations to string."""

        return enum.value

    @field_serializer("log_path", "data_path")
    def serialize_path(self, path: pathlib.Path | None) -> str | None:
        """Serialises the path."""

        return str(path) if path is not None else None

    @model_validator(mode="after")
    def validate_after(self) -> Self:
        """Performs validations after the fields have been set."""

        defaults = self.get_defaults()

        # Use internal configuration files to fill in missing fields.
        for field in [
            "min_purge_time",
            "max_purge_time",
            "min_fill_time",
            "max_fill_time",
            "max_pressure",
            "max_temperature",
            "max_temperature_increase",
        ]:
            default_value = defaults.get(field, None)
            if getattr(self, field) is None:
                setattr(self, field, default_value)

        if self.cameras == []:
            self.cameras = defaults.get("cameras", [])

        if self.valve_info is None or self.valve_info == {}:
            if "valve_info" not in self._internal_config:
                raise ValueError("No valve_info defined in the configuration file.")
            self.valve_info = self._internal_config["valve_info"]

        # At this point we should have narrowed down the interactive mode.
        if self.interactive not in [InteractiveMode.yes, InteractiveMode.no]:
            raise ValueError("Invalid interactive mode after validation.")

        # If we are not in interactive mode, we set no_prompt to True.
        if self.interactive == InteractiveMode.no:
            self.no_prompt = True

        defaults = self._internal_config.get("defaults", {})

        # Ensure the log path is set correctly.
        if self.log_path is not None:
            self.write_log = True
        elif self.write_log is False:
            self.log_path = None
        else:
            if self.log_path is not None:
                self._check_log_path(self.log_path)
            else:
                default_path = defaults.get("log_path", "./{timestamp}.log")
                self.log_path = pathlib.Path(
                    default_path.format(timestamp=self._log_basename())
                )

        # Ensure the data path is set correctly.
        if self.data_path is not None:
            self.write_data = True
        elif self.write_data is False:
            self.data_path = None
        else:
            if self.data_path is not None:
                self._check_log_path(self.data_path)
            else:
                if self.log_path is not None:
                    self.data_path = self.log_path.with_suffix(".parquet")
                else:
                    default_path = defaults.get("data_path", "./{timestamp}.parquet")
                    self.data_path = pathlib.Path(
                        default_path.format(timestamp=self._log_basename("parquet"))
                    )

        # We won't write a JSON file if we are not writing a normal log.
        if not self.write_log:
            self.write_json = False

        # We cannot purge/fill without thermistors and without fill and purge times
        # unless we are in interactive mode.
        if not self.use_thermistors and not self.interactive:
            no_purge_time = self.purge_time is None
            no_fill_time = self.fill_time is None

            if self.action == Actions.purge_fill and (no_fill_time or no_purge_time):
                raise ValueError(
                    "use_thermistors=False requires interactive mode "
                    "or defining purge and fill times."
                )

            if self.action == Actions.purge and no_purge_time:
                raise ValueError(
                    "use_thermistors=False requires interactive mode "
                    "or defining a purge time."
                )

            if self.action == Actions.fill and no_fill_time:
                raise ValueError(
                    "use_thermistors=False requires interactive mode "
                    "or defining a fill time."
                )

        # use_thermistors cannot be used with fill/purge times.
        if self.use_thermistors and (self.purge_time or self.fill_time):
            raise ValueError("use_thermistors cannot be used with purge/fill times.")

        # If we are using thermistors, require max_fill_time and max_purge_time.
        if self.use_thermistors:
            if self.max_fill_time is None or self.max_purge_time is None:
                raise ValueError(
                    "use_thermistors=True requires defining max_fill_time and "
                    "max_purge_time. This probably indicates a problem with your "
                    "configuration file."
                )

        return self

    def get_internal_config(self) -> Configuration:
        """Returns the internal configuration object."""

        return self._internal_config

    def get_defaults(self) -> dict[str, Any]:
        """Returns a dictionary with the default values for all parameters."""

        defaults = self._internal_config.get("defaults", {})

        return defaults

    def _check_log_path(self, path: pathlib.Path, extension="log") -> pathlib.Path:
        """Checks that the log path is valid."""

        if path.is_dir():
            return path / self._log_basename(extension)
        else:
            if path.exists():
                raise ValueError(f"Path {path} already exists.")

        return path

    def _log_basename(self, extension: str | None = None) -> str:
        """Generates a log basename with a timestamp."""

        extension = extension or ""

        now = datetime.datetime.now(datetime.UTC)
        iso = now.isoformat(timespec="seconds").split("+")[0]

        if extension and not extension.startswith("."):
            extension = f".{extension}"

        return f"lvmcryo_{iso}{extension}"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-11-26
# @Filename: thermistor.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import warnings
from dataclasses import dataclass
from datetime import UTC, datetime
from time import time

from typing import TYPE_CHECKING, TypedDict

from lvmopstools.devices.thermistors import read_thermistors


if TYPE_CHECKING:
    from lvmcryo.handlers.valve import ValveHandler


class ThermistorDataPoint(TypedDict):
    """A data point from the thermistor server."""

    timestamp: float
    data: dict[str, bool]


class ThermistorMonitor:
    """Queries all thermistors and stores data points. Implemented as a singleton."""

    __instance: ThermistorMonitor | None = None
    __initialised: bool = False

    def __new__(cls, interval: float = 1):
        if cls.__instance is None:
            cls.clear()

            cls.__instance = super(ThermistorMonitor, cls).__new__(cls)
            cls.__instance.__initialised = False

        return cls.__instance

    def __init__(self, interval: float = 1):
        if self.__initialised:
            return

        self.__initialised = True

        self.interval = interval

        self.data: list[ThermistorDataPoint] = []
        self._task: asyncio.Task | None = None

    def start(self):
        """Starts the monitoring task."""

        if self._task and not self._task.done():
            return

        self._task = asyncio.create_task(self._monitor())

    def stop(self):
        """Stops the monitoring task."""

        if self._task and not self._task.done():
            self._task.cancel()

    async def _monitor(self):
        """Monitors the thermistors and emits an event when new data is available."""

        while True:
            try:
                data: dict[str, bool] = await read_thermistors()
            except Exception as ee:
                warnings.warn(f"Error reading thermistors: {ee}", UserWarning)
                continue

            self.data.append({"timestamp": time(), "data": data})

            await asyncio.sleep(self.interval)

    @classmethod
    def clear(cls):
        if cls.__instance:
            cls.__instance.stop()

        cls.__instance = None
        cls.__initialised = False


@dataclass
class ThermistorHandler:
    """Reads a thermistor, potentially closing the valve if active.

    Parameters
    ----------
    valve_handler
        The `.ValveHandler` instance associated with this thermistor.
    channel
        The name of the thermistor channel to monitor.
    monitoring_interval
        The interval in seconds between thermistor checks.
    min_open_time
        The minimum valve open time. The thermistor will not be read until
        the minimum time has been reached.
    close_valve
        If ``True``, closes the valve once the thermistor is active.
    required_active_time
        Number of seconds the thermistor must be active before the valve is
        closed.
    disabled
        If ``True``, the thermistor is disabled and we will not monitor it.

    """

    valve_handler: ValveHandler
    channel: str
    monitoring_interval: float = 1.0
    min_open_time: float = 0.0
    close_valve: bool = True
    required_active_time: float = 10.0
    disabled: bool = False

    def __post_init__(self):
        self.log = self.valve_handler.log

        self.thermistor_monitor = ThermistorMonitor(interval=self.monitoring_interval)

        self.active: bool = False
        self.first_active: datetime | None = None

    async def start_monitoring(self):
        """Monitors the thermistor and potentially closes the valve."""

        # If the thermistor is not working return immediately and
        # the valve will close on a timeout.
        if self.disabled:
            self.log.warning(
                f"The thermistor for valve {self.valve_handler.valve!r} is disabled. "
                "Will not monitor it."
            )
            return

        # Unix time at which we start monitoring.
        start_time = time()
        self.thermistor_monitor.start()

        self.valve_handler.log.debug(f"Started to monitor thermistor {self.channel!r}.")

        # Unix time at which we saw the last data point.
        last_seen: float = 0

        # Unix time at which the thermistor became active.
        active_start_time: float = 0

        # Elapsed time we have been monitoring.
        elapsed_running: float = 0

        # Elapsed time the thermistor has been active
        elapsed_active: float = 0

        # How long to wait to issue a warning if ThermistorMonitor
        # is not providing new data.
        alert_seconds = int(10 * self.monitoring_interval)

        # Number of seconds since the last issued warnings.
        last_warned = -1

        while True:
            await asyncio.sleep(self.monitoring_interval)

            elapsed_running = time() - start_time

            # Check that there are measurements in the data list.
            if len(self.thermistor_monitor.data) > 0:
                # Get the last measurement.
                data_point = self.thermistor_monitor.data[-1]

                # Check that the last measurement includes the valve we are monitoring.
                if self.channel in data_point["data"]:
                    th_data = data_point["data"][self.channel]  # Boolean
                    timestamp = data_point["timestamp"]  # Unix time

                    # Check that the data is fresh.
                    if abs(timestamp - last_seen) > 0.1:
                        last_seen = timestamp

                        # If the thermistor is active, track for how long.
                        if th_data:
                            if active_start_time <= 0:
                                active_start_time = time()

                                if elapsed_running > self.min_open_time:
                                    self.log.info(
                                        f"Thermistor {self.channel!r} is active. "
                                        f"Waiting {self.required_active_time} seconds "
                                        "before closing the valve."
                                    )
                                else:
                                    self.log.warning(
                                        f"Thermistor {self.channel!r} is active but "
                                        "the minimum open time has not been reached."
                                    )

                            # Time the thermistor has been active.
                            elapsed_active = time() - active_start_time
                            if elapsed_active > self.required_active_time:
                                if not self.first_active:
                                    self.first_active = datetime.now(UTC)

                            # We require the time we have been running to
                            # be > min_open_time and the the time the thermistor
                            # has been active to be > required_active_time.
                            if (
                                active_start_time > 0
                                and elapsed_active > self.required_active_time
                                and elapsed_running > self.min_open_time
                            ):
                                # The thermistor has been active for long enough.
                                # Exit the loop and close the valve.
                                break

                        else:
                            # The thermistor is not active. Reset the active time and
                            # elapsed active time in case we had a case in which
                            # the thermistor was active for a short period and then
                            # became inactive.
                            if active_start_time > 0:
                                self.log.warning(
                                    f"Thermistor {self.channel!r} is no "
                                    "longer active. Resetting counters."
                                )

                            active_start_time = 0
                            elapsed_active = 0

            # Run some checks to be sure we are getting fresh data, but we won't
            # fail the fill if that's the case, so we only do it if there's a logger
            # to which we can report.
            if self.log is not None:
                last_seen_elapsed = time() - last_seen

                is_late = last_seen > 0 and last_seen_elapsed > alert_seconds
                never_seen = last_seen <= 0 and elapsed_running > alert_seconds

                if is_late or never_seen:
                    # Issue a warning the first time this happens
                    # and then every 30 seconds.
                    if last_warned == -1 or last_warned > 30:
                        self.log.warning(
                            f"No data from the thermistor {self.channel!r} "
                            f"in the last {last_seen_elapsed:.1f} seconds."
                        )
                        last_warned = 0
                    else:
                        last_warned += self.monitoring_interval

        self.valve_handler.log.debug(
            f"Thermistor {self.channel!r} has been active for more than "
            f"{elapsed_active:.1f} seconds."
        )
        self.active = True

        if self.close_valve:
            self.valve_handler.log.debug(
                f"Closing valve {self.valve_handler.valve!r} "
                "due to thermistor feedback."
            )
        else:
            self.valve_handler.log.warning(
                f"Thermistor {self.channel!r} is active. Calling "
                "ValveHandler.finish() but not closing the valve."
            )

        await self.valve_handler.finish(close_valve=self.close_valve)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-11-10
# @Filename: __main__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import enum
import os
import pathlib
import signal
from functools import wraps

from typing import Annotated, Optional, cast

import typer
from click.core import ParameterSource
from rich.console import Console
from typer import Argument, Option
from typer.core import TyperGroup

from lvmcryo.config import Actions, InteractiveMode, NotificationLevel, ParameterOrigin


info_console = Console()
err_console = Console(stderr=True)


DEBUG = os.environ.get("LVMCRYO_DEBUG", "").lower() not in ["", "0", "false"]


def cli_coro(
    signals=(signal.SIGHUP, signal.SIGTERM, signal.SIGINT),
    shutdown_func=None,
):
    """Decorator function that allows defining coroutines with click."""

    def decorator_cli_coro(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            loop = asyncio.get_event_loop()
            if shutdown_func:
                for ss in signals:
                    loop.add_signal_handler(ss, shutdown_func, ss, loop)

            if loop.is_running():
                return asyncio.create_task(f(*args, **kwargs))

            return loop.run_until_complete(f(*args, **kwargs))

        return wrapper

    return decorator_cli_coro


class NaturalOrderGroup(TyperGroup):
    """A Typer group that lists commands in order of definition."""

    def list_commands(self, ctx):
        return self.commands.keys()


cli = typer.Typer(
    cls=NaturalOrderGroup,
    rich_markup_mode="rich",
    context_settings={"obj": {}},
    no_args_is_help=True,
    help="CLI for LVM cryostats.",
)


def version_callback(value: bool):
    from lvmcryo import __version__

    if value:
        typer.echo(f"lcmcryo {__version__}")
        raise typer.Exit()


@cli.callback()
def main(
    version: Annotated[
        Optional[bool],
        Option(
            "--version",
            "-V",
            help="Show the version and exit.",
            is_eager=True,
            callback=version_callback,
        ),
    ] = None,
):
    """CLI for LVM cryostats."""

    pass


@cli.command("ln2")
@cli_coro()
async def ln2(
    ctx: typer.Context,
    #
    # Arguments
    #
    action: Annotated[
        Actions,
        Argument(help="Action to perform."),
    ] = Actions.purge_fill,
    cameras: Annotated[
        list[str] | None,
        Argument(
            help="Comma-separated cameras to fill. Defaults to all cameras. "
            "Ignored for [green]purge[/], [green]abort[/] and [green]clear[/]. "
            "When supplied, ACTION needs to be explicitely defined.",
        ),
    ] = None,
    #
    # General options
    #
    profile: Annotated[
        Optional[str],
        Option(
            "--profile",
            "-p",
            envvar="LVMCRYO_PROFILE",
            help="Profile to use. A list of valid profiles and parameters "
            "can be printed with lvmcryo list-profiles.",
        ),
    ] = None,
    config_file: Annotated[
        Optional[pathlib.Path],
        Option(
            dir_okay=False,
            exists=True,
            envvar="LVMCRYO_CONFIG_FILE",
            help="The configuration file. Defaults to the internal configuration file.",
        ),
    ] = None,
    dry_run: Annotated[
        bool,
        Option(
            "--dry-run",
            help="Test run the code but do not actually open/close any valves.",
        ),
    ] = False,
    interactive: Annotated[
        InteractiveMode,
        Option(
            "--interactive",
            "-i",
            envvar="LVMCRYO_INTERACTIVE",
            help="Controls whether the interactive features are shown. When "
            "--interactive auto (the default), interactivity is determined "
            "based on console mode.",
        ),
    ] = InteractiveMode.auto,
    no_prompt: Annotated[
        bool,
        Option(
            "--no-prompt",
            "-y",
            envvar="LVMCRYO_NO_PROMPT",
            help="Does not prompt the user to finish or abort a purge/fill. "
            "Prompting is required if --fill-time or --purge-time are not provided "
            "and thermistors are not used.",
        ),
    ] = False,
    with_traceback: Annotated[
        bool,
        Option(
            "--with-traceback",
            envvar="LVMCRYO_DEBUG",
            help="Show the full traceback in case of an error. If not set, only "
            "the error message is shown. The full traceback is always logged "
            "to the log file.",
            show_default=False,
        ),
    ] = False,
    clear_lock: Annotated[
        bool,
        Option(
            "--clear-lock",
            help="Clears the lock file if it exists before carrying out the action.",
        ),
    ] = False,
    #
    # Purge and fill options
    #
    use_thermistors: Annotated[
        bool,
        Option(
            " /--no-use-thermistors",
            envvar="LVMCRYO_USE_THERMISTORS",
            help="Use thermistor values to determine purge/fill time.",
            rich_help_panel="Purge and fill options",
        ),
    ] = True,
    require_all_thermistors: Annotated[
        bool,
        Option(
            "--require-all-thermistors",
            envvar="LVMCRYO_REQUIRE_ALL_THERMISTORS",
            help="If set, waits until all thermistors have activated before closing "
            "any of the valves. This prevents over pressures in the cryostats when "
            "only some of the valves are open. Ignored if --no-use-thermistors "
            "is used.",
            rich_help_panel="Purge and fill options",
        ),
    ] = False,
    check_pressures: Annotated[
        bool,
        Option(
            " /--no-check-pressures",
            help="Aborts purge/fill if the pressure of any cryostat "
            "is above the limit.",
            rich_help_panel="Purge and fill options",
        ),
    ] = True,
    check_temperatures: Annotated[
        bool,
        Option(
            " /--no-check-temperatures",
            help="Aborts purge/fill if the temperature of a fill cryostat "
            "is above the limit.",
            rich_help_panel="Purge and fill options",
        ),
    ] = True,
    check_o2_sensors: Annotated[
        bool,
        Option(
            " /--no-check-o2-sensors",
            help="Aborts purge/fill if the oxygen sensor reads below the limit.",
            rich_help_panel="Purge and fill options",
        ),
    ] = True,
    max_pressure: Annotated[
        Optional[float],
        Option(
            help="Maximum cryostat pressure if --check-pressure. "
            "Defaults to internal value.",
            show_default=False,
            rich_help_panel="Purge and fill options",
        ),
    ] = None,
    max_temperature: Annotated[
        Optional[float],
        Option(
            help="Maximum cryostat temperature if --check-temperature. "
            "Defaults to internal value.",
            show_default=False,
            rich_help_panel="Purge and fill options",
        ),
    ] = None,
    purge_time: Annotated[
        Optional[float],
        Option(
            help="Purge time in seconds. If --use-thermistors, this is "
            "used as the maximum purge time.",
            rich_help_panel="Purge and fill options",
        ),
    ] = None,
    min_purge_time: Annotated[
        Optional[float],
        Option(
            help="Minimum purge time in seconds. Defaults to internal value.",
            show_default=False,
            rich_help_panel="Purge and fill options",
        ),
    ] = None,
    max_purge_time: Annotated[
        Optional[float],
        Option(
            help="Maximum purge time in seconds. Defaults to internal value.",
            show_default=False,
            rich_help_panel="Purge and fill options",
        ),
    ] = None,
    fill_time: Annotated[
        Optional[float],
        Option(
            help="Fill time in seconds. If --use-thermistors, this is "
            "used as the maximum fill time.",
            rich_help_panel="Purge and fill options",
        ),
    ] = None,
    min_fill_time: Annotated[
        Optional[float],
        Option(
            help="Minimum fill time in seconds. Defaults to internal value.",
            show_default=False,
            rich_help_panel="Purge and fill options",
        ),
    ] = None,
    max_fill_time: Annotated[
        Optional[float],
        Option(
            help="Maximum fill time in seconds. Defaults to internal value.",
            show_default=False,
            rich_help_panel="Purge and fill options",
        ),
    ] = None,
    #
    # Notification options
    #
    notify: Annotated[
        bool,
        Option(
            "--notify",
            envvar="LVMCRYO_NOTIFY",
            help="Sends a notification when the action is completed. "
            "In not set, --slack and --email are ignored. "
            "Notifications are not sent for the [green]abort[/] "
            "and [green]clear-lock[/] actions.",
            rich_help_panel="Notifications",
            show_default=True,
        ),
    ] = False,
    slack: Annotated[
        bool,
        Option(
            help="Send a Slack notification when the action is completed.",
            rich_help_panel="Notifications",
        ),
    ] = True,
    email: Annotated[
        bool,
        Option(
            help="Send an email notification when the action is completed.",
            rich_help_panel="Notifications",
        ),
    ] = True,
    email_level: Annotated[
        NotificationLevel,
        Option(
            envvar="LVMCRYO_EMAIL_LEVEL",
            help="Send an email notification only on error or always.",
            case_sensitive=False,
            rich_help_panel="Notifications",
        ),
    ] = NotificationLevel.error,
    #
    # Logging options
    #
    quiet: Annotated[
        bool,
        Option(
            "--quiet",
            "-q",
            help="Disable logging to stdout.",
            rich_help_panel="Logging",
        ),
    ] = False,
    verbose: Annotated[
        bool,
        Option(
            "--verbose",
            "-v",
            help="Outputs additional information to stdout.",
            rich_help_panel="Logging",
        ),
    ] = False,
    write_log: Annotated[
        bool,
        Option(
            "--write-log",
            "-S",
            envvar="LVMCRYO_WRITE_LOG",
            help="Saves the log to a file.",
            rich_help_panel="Logging",
        ),
    ] = False,
    log_path: Annotated[
        Optional[pathlib.Path],
        Option(
            exists=False,
            help="Path where to save the log file. Implies --write-log. "
            "Defaults to the current directory.",
            rich_help_panel="Logging",
        ),
    ] = None,
    write_json: Annotated[
        bool,
        Option(
            "--write-json",
            envvar="LVMCRYO_WRITE_JSON",
            help="Saves the log in JSON format. Uses the same path as the file log. "
            "Ignored if --write-log is not passed.",
            rich_help_panel="Logging",
        ),
    ] = True,
    #
    # Post-fill data options
    #
    write_data: Annotated[
        bool,
        Option(
            "--write-data",
            envvar="LVMCRYO_WRITE_DATA",
            help="Saves cryostat pressures and temperatures taken during purge/fill "
            "to a parquet file.",
            rich_help_panel="Post-fill data logging",
        ),
    ] = False,
    data_path: Annotated[
        Optional[pathlib.Path],
        Option(
            envvar="LVMCRYO_DATA_PATH",
            exists=False,
            help="Path where to save the data. Implies --write-data. "
            "Defaults to a path relative to the log path.",
            rich_help_panel="Post-fill data logging",
        ),
    ] = None,
    data_extra_time: Annotated[
        float,
        Option(
            envvar="LVMCRYO_DATA_EXTRA_TIME",
            help="Additional time to take cryostat data after the "
            "action has been completed. The command will not complete until "
            "data collection has finished.",
            rich_help_panel="Post-fill data logging",
        ),
    ] = 60,
):
    """Handles LN2 purges and fills.

    Allowed actions are: [bold blue]purge-and-fill[/] which executes a purge followed
    by a fill, [bold blue]purge[/] which only purges, [bold blue]fill[/] which only
    fills, [bold blue]abort[/] which aborts an ongoing purge/fill by closing all the
    valves, and [bold blue]clear[/] which removes the lock.

    """

    from sdsstools.logger import get_logger

    from lvmcryo.runner import LN2RunnerError, ln2_runner
    from lvmcryo.tools import LockExistsError

    log = get_logger("lvmcryo", use_rich_handler=True)
    log.setLevel(5)

    try:
        # Determine parameter origins. The config uses this information to
        # decide whether a profile parameter should be overridden or not.
        param_origin: dict[str, ParameterOrigin | None] = {}
        for param in ctx.params:
            source = ctx.get_parameter_source(param)
            match source:
                case ParameterSource.COMMANDLINE:
                    param_origin[param] = ParameterOrigin.COMMAND_LINE
                case ParameterSource.DEFAULT:
                    param_origin[param] = ParameterOrigin.DEFAULT
                case ParameterSource.ENVIRONMENT:
                    param_origin[param] = ParameterOrigin.ENVVAR
                case _:
                    param_origin[param] = None

        await ln2_runner(
            action=action,
            cameras=cameras or [],
            config_file=config_file,
            dry_run=dry_run,
            clear_lock=clear_lock,
            with_traceback=with_traceback,
            interactive=interactive,
            no_prompt=no_prompt,
            notify=notify,
            slack=slack,
            email=email,
            email_level=email_level,
            use_thermistors=use_thermistors,
            require_all_thermistors=require_all_thermistors,
            check_pressures=check_pressures,
            check_temperatures=check_temperatures,
            check_o2_sensors=check_o2_sensors,
            max_pressure=max_pressure,
            max_temperature=max_temperature,
            purge_time=purge_time,
            min_purge_time=min_purge_time,
            max_purge_time=max_purge_time,
            fill_time=fill_time,
            min_fill_time=min_fill_time,
            max_fill_time=max_fill_time,
            quiet=quiet,
            verbose=verbose,
            write_log=write_log,
            write_json=write_json,
            log_path=log_path,
            write_data=write_data,
            data_path=data_path,
            data_extra_time=data_extra_time,
            profile=profile,
            log=log,
            __parameter_origin=param_origin,
        )

    except LockExistsError:
        err_console.print(
            "[red]Lock file already exists.[/] Another instance may be "
            "performing an operation. Use [green]lvmcryo ln2 clear-lock[/] to remove "
            "the lock."
        )
        raise typer.Exit(1)

    except Exception as err:
        # Hide the traceback unless with_traceback or DEBUG is set, but always log it
        # to the file, if we are doing that.
        log.sh.setLevel(10000)

        log.exception("Error raised in LN2 runner.", exc_info=err)
        err_console.print(f"[red]LN2 operation failed:[/] {err}")

        if (isinstance(err, LN2RunnerError) and err.propagate) or DEBUG:
            raise

        raise typer.Exit(1)

    return typer.Exit(0)


async def _close_valves_helper():
    """Closes all the outlets."""

    from lvmcryo.handlers.valve import close_all_valves

    try:
        await close_all_valves()
        info_console.print("[green]All valves closed.[/]")
    except Exception as err:
        info_console.print(f"[red]Error closing valves:[/] {err}")
        raise typer.Exit(1)

    return typer.Exit(0)


@cli.command("list-profiles")
def list_profiles(
    config_file: Annotated[
        Optional[pathlib.Path],
        Option(
            dir_okay=False,
            exists=True,
            envvar="LVMCRYO_CONFIG_FILE",
            help="The configuration file. Defaults to the internal configuration file.",
        ),
    ] = None,
):
    """Lists the available profiles."""

    from lvmcryo.config import get_internal_config

    internal_config = get_internal_config(config_file=config_file)
    profiles = internal_config["profiles"]

    for profile in profiles:
        info_console.print(profile, style="bold")
        info_console.print(profiles[profile])
        print()


@cli.command("clear-lock")
@cli_coro()
async def clear_lock(
    wait: Annotated[
        bool,
        Option(
            "--wait/--no-wait",
            help="Waits a few seconds after removing the lock file to allow other "
            "processes to stop.",
            show_default=True,
        ),
    ] = True,
    wait_delay: Annotated[
        float,
        Option(
            "--wait-delay",
            help="Time in seconds to wait after removing the lock file.",
            show_default=True,
        ),
    ] = 10.0,
):
    """Clears the lock file if it exists."""

    from lvmcryo.config import get_internal_config

    config = get_internal_config()
    lockfile_path = pathlib.Path(config.get("lockfile", "/data/lvmcryo.lock"))

    if lockfile_path.exists():
        lockfile_path.unlink()
        info_console.print("[green]Lock file removed.[/]")

        if wait:
            wait_delay = int(wait_delay)
            info_console.print(
                f"[gray]Waiting {wait_delay} seconds for other processes to stop.[/]"
            )
            await asyncio.sleep(wait_delay)

    else:
        info_console.print("[yellow]No lock file found.[/]")

    return typer.Exit(0)


@cli.command("abort")
@cli_coro()
async def abort():
    """Aborts an ongoing purge/fill and closes all the valves.

    Equivalent of calling close-valves and clear-lock.

    """

    await close_valves()
    await clear_lock(wait=True)


@cli.command("close-valves")
@cli_coro()
async def close_valves():
    """Closes all solenoid valves."""

    return await _close_valves_helper()


@cli.command("ion")
@cli_coro()
async def ion(
    cameras: Annotated[
        list[str] | None,
        Argument(
            help="List of cameras to handle. Defaults to all cameras.",
            case_sensitive=False,
            show_default=False,
        ),
    ] = None,
    on: Annotated[
        Optional[bool],
        Option(
            "--on/--off",
            help="Turns the ion pump on or off. If not provided, "
            "the current status of the ion pump is returned.",
            show_default=False,
        ),
    ] = None,
    skip_pressure_check: Annotated[
        bool,
        Option(
            "--skip-pressure-check",
            help="Skips the pressure check when turning the ion pump on.",
        ),
    ] = False,
):
    """Controls the ion pumps.

    Without ``--on`` or ``--off``, returns the current status of the ion pump.
    A list of space-separated cameras can be provided.

    """

    import math
    import warnings

    from lvmopstools.devices.ion import read_ion_pumps, toggle_ion_pump
    from lvmopstools.devices.specs import Spectrographs, spectrograph_pressures

    from lvmcryo.config import get_internal_config

    cameras = list(map(lambda x: x.lower(), cameras)) if cameras else ["all"]

    if "all" in cameras:
        cameras = ["all"]

    if on is None:
        query_cameras = cameras if cameras != ["all"] else None
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            status = await read_ion_pumps(query_cameras)

        failed_cameras: list[str] = []
        for camera, data in status.items():
            if data["on"] is None:
                failed_cameras.append(camera)

        if len(failed_cameras) > 0:
            info_console.print(
                "[yellow]Failed getting ion pump status for cameras:[/] "
                f"{', '.join(failed_cameras)}"
            )

        status_pretty = {}
        for camera, data in status.items():
            if camera not in failed_cameras:
                pressure = data["pressure"]
                status_pretty[camera] = {
                    "pressure": float(f"{pressure:.2e}")  # Truncate to 2 decimals
                    if pressure is not None
                    else math.nan,
                    "on": data["on"],
                }

        info_console.print(status_pretty, width=80)
        return

    error: bool = False

    if cameras == ["all"]:
        config = get_internal_config()
        cameras = config["defaults"]["cameras"]

    if cameras is None:
        err_console.print("[red]No cameras were specified.[/]")
        raise typer.Exit(1)

    for camera in cameras:
        try:
            if on is True and not skip_pressure_check:
                spec = f"sp{camera[-1]}"
                pressures = await spectrograph_pressures(cast(Spectrographs, spec))
                cam_pressure = pressures.get(camera, None)
                if cam_pressure is None:
                    raise ValueError(f"Could not get pressure for {camera!r}.")

                if cam_pressure > 1e-4:
                    raise ValueError(
                        f"Camera {camera} has a pressure of {cam_pressure:.2e} "
                        "Torr. If you want to turn it on use --skip-pressure-check."
                    )

            await toggle_ion_pump(camera, on)
        except Exception as err:
            info_console.print(
                f"[yellow]Error handling ion pump for camera {camera}:[/] {err}"
            )
            error = True

    return typer.Exit(error)


class AutoFillAction(enum.Enum):
    """Auto-fill actions."""

    enable = "enable"
    disable = "disable"
    restart = "restart"


@cli.command("auto-fill")
def auto_fill(
    action: Annotated[
        AutoFillAction,
        Argument(
            help="Auto-fill action.",
            case_sensitive=False,
        ),
    ],
):
    """Enables or disables the auto-fill procedure."""

    from lvmcryo.tools import run_command

    CRONJOB_PATH = "/home/sdss5/config/kube/cronjobs/ln2fill_2_fills.yml"

    try:
        if action == AutoFillAction.enable:
            run_command(
                ["kubectl", "apply", "-f", CRONJOB_PATH],
                output_on_error=True,
                raise_on_error=True,
            )
            info_console.print("[green]Auto-fill enabled.[/]")
        elif action == AutoFillAction.disable:
            run_command(
                ["kubectl", "delete", "-f", CRONJOB_PATH],
                output_on_error=True,
                raise_on_error=True,
            )
            info_console.print("[yellow]Auto-fill disabled.[/]")
        elif action == AutoFillAction.restart:
            run_command(
                ["kubectl", "replace", "--force", "-f", CRONJOB_PATH],
                output_on_error=True,
                raise_on_error=True,
            )
            info_console.print("[green]Auto-fill restarted.[/]")
    except FileNotFoundError:
        err_console.print(
            "[red]kubectl not found. Is Kubernetes installed on this system?[/]"
        )
        raise typer.Exit(1)
    except RuntimeError:
        # Just exit with error. run_commnad should have output the error message.
        raise typer.Exit(1)


if __name__ == "__main__":
    cli()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2026-01-14
# @Filename: app.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import base64
import logging
import os
from contextlib import asynccontextmanager

from typing import Annotated, AsyncIterator

from fastapi import Body, FastAPI, Query
from pydantic import BaseModel, Field

from lvmcryo.config import get_internal_config
from lvmcryo.handlers.valve import close_all_valves
from lvmcryo.runner import clear_lock as clear_lock_helper
from lvmcryo.runner import ln2_runner
from lvmcryo.tools import lockfile_exists


logger = logging.getLogger("uvicorn.error")


@asynccontextmanager
async def lifespan(_: FastAPI) -> AsyncIterator[None]:
    yield


app = FastAPI(swagger_ui_parameters={"tagsSorter": "alpha"}, lifespan=lifespan)


@app.get("/ping", summary="Health check endpoint.")
async def ping():
    return {"status": "ok"}


@app.get("/filling", summary="Check if LN2 fill operation is ongoing.")
async def filling():
    """Checks if an LN2 fill operation is ongoing."""

    return {"filling": lockfile_exists()}


class ManualFillRequestBody(BaseModel):
    password: Annotated[
        str | None, Field(description="Password to authorize the LN2 fill.")
    ] = None


@app.post("/manual-fill", summary="Starts a manual LN2 fill operation.")
async def manual_fill(
    data: Annotated[
        ManualFillRequestBody | None, Body(description="Body of the request")
    ] = None,
    clear_lock: Annotated[
        bool,
        Query(
            description="If true, clears any existing lock before starting the fill."
        ),
    ] = False,
):
    """Starts a manual LN2 purge and fill operation.

    The route returns immediately after starting the operation in the background.

    """

    password = data.password if data is not None else None

    config = get_internal_config()
    require_password = config["server.require_password"]

    if require_password is True or require_password is None:
        password_b64 = os.environ.get("LVMCRYO_FILL_PASSWORD", None)

        if password_b64 is None:
            return {"result": False, "error": "Fill password not available."}

        password_bytes = password_b64.encode("utf-8")
        password_pt = base64.b64decode(password_bytes).decode("utf-8")

        if password != password_pt:
            return {"result": False, "error": "Invalid password."}

    tasks: list[asyncio.Task] = []

    if lockfile_exists():
        if not clear_lock:
            return {
                "result": False,
                "error": "Lock file exists. LN2 fill already in process.",
            }

    tasks.append(
        ln2_runner(
            profile="manual-fill",
            clear_lock=clear_lock,
            quiet=True,
        )
    )

    asyncio.gather(*tasks)

    return {"result": True, "error": None}


@app.get("/abort", summary="Abort LN2 fill operations and releases the lock.")
async def abort_ln2_fill(
    wait: Annotated[
        bool,
        Query(
            description="Wait for the operation to complete. "
            "Otherwise returns immediately and runs in the background."
        ),
    ] = False,
):
    """Aborts any ongoing LN2 fill operation and releases the lock."""

    async def _abort():
        await close_all_valves()
        await clear_lock_helper(wait=True)

    abort_task = asyncio.create_task(_abort())

    if wait:
        await abort_task

    return {"result": True}
